/*
	brum - sort e-mail over IMAP utility	
	Written by Anicka <anicka@anicka.net> in 2008

	Can be used according to terms of GPL version 2.
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <poll.h>
#include <gnutls/gnutls.h>
#include <netdb.h>
#include <unistd.h>
#include <pcre.h>
#include <assert.h>
#define _GNU_SOURCE
#include <getopt.h>

#include "common.h"
#include "lists.h"
#include "parseconf.h"

#define TLS_CHECK(name) if (err < 0) die(#name " failed: %s", gnutls_strerror(err))

int port;
int secure;
int discard_security;
int sock;
char* cert;
static gnutls_certificate_credentials_t cert_cred;
gnutls_session_t sess;
char* login, * password, * cert;
char* default_mailbox;

void
tls_init(void)
{
	int err;

	gnutls_global_init();

	err = gnutls_certificate_allocate_credentials(&cert_cred);	
	
	TLS_CHECK(gnutls_certificate_allocate_credentials);
	if (discard_security)
		return;

 	err = gnutls_certificate_set_x509_trust_file(cert_cred, cert, GNUTLS_X509_FMT_PEM);
  	if (!err)
		die("No CA certificate found");
	if (err < 0)
		die("Unable to load X509 trust file: %s", gnutls_strerror(err));

}

/* note: only way to decide whether connection is secure or not is done
   according to port, right now. Should be configurable. FIXME */
int
xread(char* buf, int len)
{
	int err;

	if (secure){
		err=gnutls_record_recv(sess,buf,len); 
		if (err<0 && errno != EAGAIN)
			die("gnutls_record_recv: %m");
	} else {
		err=read(sock,buf,len);
		if (err<0 && errno != EAGAIN)
			die("read: %m");
	}
	return err;
}

int
xwrite(char* buf, int len)
{
	int err;

	if (secure){
		err=gnutls_record_send(sess,buf,len); 
		TLS_CHECK(gnutls_record_send);
	} else {
		err=write(sock,buf,len);
		if (err<0 && errno != EAGAIN)
			die(strerror(errno));
	}
	return err;
}

void
xclose(void)
{
	int mode = fcntl(sock, F_GETFL);
	if (mode < 0 || fcntl(sock, F_SETFL, mode & ~O_NONBLOCK) < 0)
		die("Cannot reset non-blocking mode: %m");

	if (secure){
		gnutls_bye(sess, GNUTLS_SHUT_RDWR);
 	 	gnutls_deinit(sess);
	}

	close(sock);
}

char readbuf[BUFSIZE], writebuf[BUFSIZE];
int read_pos;
int write_pos, write_len;

/* structures for storing commands meant to be sent to server and collected replies */

struct reply {
	struct node car;
	int content_bytes;
	int len;
	char line[1];
};

struct command {
	struct node car;
	int id;
	struct list response;
	int success;
	char command[1];
};

struct list cmds_to_write, cmds_waiting, cmds_done; 
struct list current_reply;
struct list* fetched_inbox;
struct list parsed_inbox; // list_first is the current one

void
init_cmds_storage(void)
{
	list_init(&cmds_to_write);
	list_init(&cmds_waiting);
	list_init(&cmds_done);
	list_init(&current_reply);
	list_init(&parsed_inbox);
}

/* search for particular id in the list of commands waiting for responses */
struct command*
find_id(struct list* l, int id)
{
	struct command* p, * found;

	found=NULL;
	LIST_FOREACH(p,l){
		if (p->id==id){
			found=p;
			break;
		}
	}
	
	return found;
}

void
free_reply(struct list* reply)
{
	struct reply* tmp;

	while (tmp = list_del_first(reply))
		free(tmp);
}

void
copy_reply(struct list* dst, struct list* src)
{
	*dst=*src;
	dst->head.prev->next=&dst->head;
	dst->head.next->prev=&dst->head;
}

/*
convert the line received from server to the part of the reply and append to original command
*/
int awaited_data;

void
process_line(char* line)
{
	struct reply* new;
	char* realcmd;
	int realcmdlen, len;

	len=strlen(line);

	/* sanity check */
	assert(awaited_data>=0);

	/* is it continuation of a communication? */
	if (awaited_data){
		/* we got part of awaited string */
		awaited_data-=(len+1);

		new=xmalloc(sizeof(struct reply)+len);
		strcpy(new->line,line);	
		new->len=len;	
		/* just save the data and we are done for now */	
		list_add_last(&current_reply,&new->car);
		return;

	/*it is not: */
	} else {
		for(realcmd=line; *realcmd!=' ' && *realcmd!=0; realcmd++);
		if (!*realcmd)
			return; /* looks like empty line */
		realcmd++; /* should point behind the first space */
	
		realcmdlen=strlen(realcmd);
		new=xmalloc(sizeof(struct reply)+realcmdlen);
		new->len=realcmdlen;
		strcpy(new->line,realcmd);	
	}
	/*add a new line of communication without leading mark, if first */
	list_add_last(&current_reply,&new->car);
	
	/*does the string with data follow? */
	if (len>1 && line[len-2]=='}'){
		int i, res;

		for (i=len-2;i>=0;i--){
			if (line[i]=='{')
				break;
		}
		if (line[i]!='{')
			die("Cannot find starting { in server reply");
		res=sscanf(line+i+1,"%d",&awaited_data);
		if (res<=0)
			die("Cannot read how much data is expected");
		new->content_bytes=awaited_data;	
	}

	/* is reply ending here? */
	if (line[0]!='*'){
		int id;
		struct command* found;

		int res=sscanf(line,"%x",&id);
		if (res<=0)
			die("Server sent us some mess instead of request id");
		found=find_id(&cmds_waiting,id);
		if (!found){
			free_reply(&current_reply);
			return;
		}

		if (realcmdlen>=2 && realcmd[0]=='O' && realcmd[1]=='K')
			found->success=1;
		else
			found->success=0;
				
		copy_reply(&found->response,&current_reply);
		list_init(&current_reply);
		list_del_node(&found->car);
		list_add_last(&cmds_done,&found->car);	

	}

}

/*
loop for communication of server. Right now, it just sends all the waiting commands to the server
and stops after collecting all the replies
*/
void 
io_loop(void)
{
	struct pollfd pfd = { .fd = sock };
	while (!list_is_empty(&cmds_to_write) || !list_is_empty(&cmds_waiting)) {
		pfd.events = POLLIN;
		if (write_pos >= write_len) {
			if (!list_is_empty(&cmds_to_write)){
				struct command* current_command;

				current_command=list_del_first(&cmds_to_write);
				list_add_last(&cmds_waiting,&current_command->car);
				//FIXME
				write_pos=0;
				write_len=sprintf(writebuf,"%s\r\n",
								current_command->command);
			}		
		}
		if (write_pos < write_len)
			pfd.events |= POLLOUT | POLLERR;
		if (poll(&pfd, 1, -1) < 0)
			die("poll: %m");
		if (pfd.revents & POLLERR)
			die("poll reported error, why?");
		if (pfd.revents & POLLOUT) {
			int cnt = xwrite(writebuf + write_pos, write_len - write_pos);
			if (debug>1)
				write(1,writebuf+write_pos,cnt);
			if (cnt > 0)
				write_pos += cnt;
		}
		if (pfd.revents & POLLIN) {
			int cnt = xread(readbuf + read_pos, BUFSIZE - read_pos);
			if (cnt >= 0) {
				char *line_start = readbuf;
				char *scan_start = readbuf + read_pos;
				char *scan_end = scan_start + cnt;
				char *c;
				while (scan_start < scan_end &&
				       (c = memchr(scan_start, '\n', scan_end - scan_start))) {
					*c++ = 0;
					if (debug>1) /* FIXME: cope with \r in debug output */
						printf(" got line: <%s>\n", line_start);
					process_line(line_start);
					scan_start = line_start = c;
				}
				read_pos = scan_end - line_start;
				if (line_start > readbuf)
					memmove(readbuf, line_start, read_pos);
				else if (read_pos >= BUFSIZE)
					die("Too long a line from the server");
			}
		}
	}
}

int curtag;

/* convert the tag and the string into a command, add to waiting commands, call io_loop 
note: in fact, by using this function we are transmitting commands one by one :( FIXME
*/
void
write_command(int tag, char* msg)
{
	struct command* new;
	int len;
	char buf[BUFSIZE];	

	len=sprintf(buf,"%x %s",tag,msg);
	if (len>BUFSIZE-2) /* leave space for \r\n */
		die("Too long command: %x %s",tag,msg);	
		
	new=xmalloc(len+sizeof(struct command));
	strcpy(new->command,buf);
	new->id=tag;	

	list_add_last(&cmds_to_write,&new->car);
	io_loop();	
}
/* collect response from the server in done commands list */
struct command*
get_response(int tag)
{
	struct command* p;

	/* believe it is somewhere in this list */
	LIST_FOREACH(p,&cmds_done){ 
		if (p->id==tag)
			break;
				
	}
	list_del_node(&p->car);

	return p;
}

/* FIXME: we know only one way of authentication, if it is not supported, we just die */
void
authenticate(char* name, char* pwd)
{
	struct command* response;
	char cmd[BUFSIZE];
	int tag=curtag++;

	sprintf(cmd,"LOGIN %s %s",name,pwd);
	write_command(tag,cmd);
	response=get_response(tag);

	/* believe we should not get NULL */
	if (!response->success)
		die("Authentication failed");	

	free_reply(&response->response);
	free(response);
}

void
logout(void)
{
	struct command* response;
	int tag=curtag++;

	write_command(tag,"CLOSE");
	response=get_response(tag);
	if (!response->success)
		die("Close failed");	

	free_reply(&response->response);
	free(response);

	tag=curtag++;
	write_command(tag,"LOGOUT");
	response=get_response(tag);
	if (!response->success)
		die("Logout failed");	

	free_reply(&response->response);
	free(response);
}

/*
we just ignore all the output :(
FIXME: we should not attempt to modify the INBOX if it is read only and die right away
*/
void
select_mailbox(char* mailbox)
{
	struct command* response;
	char cmd[BUFSIZE];
	int tag=curtag++;

	sprintf(cmd,"SELECT %s",mailbox);
	write_command(tag,cmd);
	response=get_response(tag);

	if (!response->success)
		die("Mailbox %s selecting failed",mailbox);	

	free_reply(&response->response);
	free(response);
}

/* get all the messages in INBOX, logout and exit if empty
FIXME: some exit routine would be fine, we are calling the same three functions two times
*/
void
fetch_inbox(void)
{
        struct command* response;
        char cmd[BUFSIZE];
        int tag=curtag++;

        sprintf(cmd,"FETCH 1:* (UID BODY.PEEK[HEADER])");
        write_command(tag,cmd);
        response=get_response(tag);

        /* No problem, we just exit here */
        if (!response->success){
        	logout();
		xclose(); 
		if (debug)
			printf("INBOX empty, exiting\n");
		exit(0);
	}
        fetched_inbox=&response->response;
}

/*
 these two structures are used for storing fetched and parsed messages
 FIXME: we work only with headers right now
*/
struct message {
	struct node car;
	int uid;
	int content_bytes;
	struct list headers;
	char* target_mailbox;
};

struct header {
	struct node car;
	char line[1];
};

/*
FIXME: works for me just by the chance. To do the difference we seem to need a parses, unfortunately. So later.
*/
int
get_uid(char* line)
{
	char* pt;
	int uid, res;

	pt=strstr(line,"UID ");
	if (!pt){
		if (debug>4)
			printf("No UID from %s",line);
		return 0; /* filthy hack */	
	}
	res=sscanf(pt+4,"%d",&uid);
	if (res<=0)
		die("Could not parse message UID from line %s",line);
	return uid;
}

void
cat_header_lines(struct header* start, struct header* end, int size)
{
	struct header *newline, *pt, *tmp;
	int offset=0;

	newline=xmalloc(sizeof(struct reply)+size);

	list_add_after(&end->car,&newline->car);

	pt=start;
	while(pt!=newline){
		tmp=pt;
		strcpy(newline->line+offset,pt->line);
		offset+=strlen(pt->line);
		pt=list_next(fetched_inbox,(struct node*)pt);
		list_del_node(&tmp->car);
		free(tmp);
	}
	assert(offset==size);	
}

/*
We are kind of guessing where the headers are stored. We also suppose to get the reply in correct order. It seems to work but it does not have to. FIXME but we need the damn parser.
*/
void
parse_inbox(void)
{
	struct reply* pt;
	struct message* pm;
	struct header* ph, *pe, *ps;
	int left=0;

	if (debug>4)
		printf("Starting to parse INBOX\n");

	LIST_FOREACH(pt, fetched_inbox){
		/* just part of the header */
		if (left){
			struct header* new;

			left-=(pt->len+1);
			assert(left>=0);
			new=xmalloc(sizeof(struct header)+pt->len);
			memcpy(new->line, pt->line,pt->len+1);
			struct message* curr=list_first(&parsed_inbox);
			list_add_last(&curr->headers,&new->car);
			if (debug>4)
				printf("\t%s\n",pt->line);

		/* message starts */
		} else { 
			struct message* new;
			int uid;

			uid=get_uid(pt->line);
			if (!uid)
				continue;
			new=xmalloc(sizeof(struct message));
			list_init(&new->headers);
			new->uid=uid;
			new->target_mailbox=default_mailbox;
			if (pt->content_bytes)
				left=new->content_bytes=pt->content_bytes;
			list_add_first(&parsed_inbox,&new->car);			
			if (debug>4)
				printf("New message with uid %d: %s\n",uid,pt->line);
		}
	}
	if (debug>4)
		printf("Starting to concat headers\n");

	LIST_FOREACH(pm,&parsed_inbox){
		ps=(list_first(&pm->headers));
		pe=NULL;
		int size=0;
		LIST_FOREACH(ph,&pm->headers){
			if (ph->line[0]==' ' || ph->line[0]=='\t'){
				pe=ph;
				size+=strlen(ph->line);
			}
			else {
				if (pe)
					cat_header_lines(ps,pe,size);
				size=strlen(ph->line);
				pe=NULL;
				ps=ph;
			}	
		}
		if (pe)
			cat_header_lines(ps,pe,size);
	}
}

#define OVECCOUNT 3
int see_from;
/* just match all the headers with all the regexp rules one by one. Program calls are not implemented yet, TODO */
void
apply_rules(void)
{
	struct message* pm;
	struct header* ph;
	struct rule* prul;
	struct recipe* prec;

	LIST_FOREACH(pm,&parsed_inbox){
		LIST_FOREACH(prec,&recipe_list){
			int recipe_applied=1;
				LIST_FOREACH(prul,&prec->rules){
					int rule_applied=0;
					LIST_FOREACH(ph,&pm->headers){
						int ovector[OVECCOUNT];
						/*right now, we just do not care for program calls */
						if (prul->type==CALL){
							rule_applied++;
							break;
						}
						int res=pcre_exec(prul->rule.r.compreg,NULL,
								ph->line, strlen(ph->line),0,0,
								ovector,OVECCOUNT);
						if (res>0){
							rule_applied++;
							break;	
						}
					}
					if (!rule_applied){
						recipe_applied=0;
						break;
					}
				}
				if (recipe_applied){
					pm->target_mailbox=prec->mailbox;
					break;	
			}
		}
		/* a nice list of newly sorted messages for -v */
		if (debug){
			struct header* ph;
			int ok=0;
			
			LIST_FOREACH(ph,&pm->headers){
                                if (!strncmp(ph->line,"Message-ID:",strlen("Message-ID:")) || 
					!strncmp(ph->line,"Message-Id:",strlen("Message-Id:"))){
					ok=1;
                        	        break;
				}
			}
			printf("%s: %s\n",pm->target_mailbox,ok?ph->line+12:"unknown msgid");
		}

		/* we want to see a sender */
		if (see_from){
			struct header* ph;
			int ok=0;
			
			LIST_FOREACH(ph,&pm->headers){
                                if (!strncmp(ph->line,"From:",strlen("From:"))){
					ok=1;
                        	        break;
				}
			}
			if (ok)
				printf("%s%s\n",debug?"\t":"",ph->line+6);
		}
	}
}

/*
Copy the message to its destination folder and if OK, delete it. We do not attempt to create folders, we are not M$.
*/
void
move_messages(void)
{
	struct command* response;
        char cmd[BUFSIZE];
	struct message* pm;
        int tag;
	int ok;

	LIST_FOREACH(pm,&parsed_inbox){
        	tag=curtag++;
		sprintf(cmd,"UID COPY %d %s",pm->uid,pm->target_mailbox);
        	write_command(tag,cmd);
        	response=get_response(tag);
        	ok=response->success;
		
       		free_reply(&response->response);
       		free(response);

		if (!ok){
			fprintf(stderr,"Could not copy message to mailbox %s, leaving in INBOX\n",pm->target_mailbox);
			continue; /* We could not copy this message, so we try next. */
		}
        	tag=curtag++;
		sprintf(cmd,"UID STORE %d +FLAGS (\\Deleted)",pm->uid);
        	write_command(tag,cmd);
        	response=get_response(tag);
        	ok=response->success;
		
       		free_reply(&response->response);
       		free(response);
		if (!ok)
			fprintf(stderr,"Could not delete a message from INBOX\n");
	}

}
/*
authfile has a very crude syntax :(
TODO: move it to the parse file
*/
void
parse_authfile(char *authfile)
{
	FILE* fr;
	char line[BUFSIZE];
	int nl=0;

	if (!(fr=fopen(authfile,"r")))
		die("Cannot open authfile");

	while (fgets(line,sizeof(line),fr)){
		int len=strlen(line)-1;
                if (line[len]=='\n')
                        line[len]=0;
                else
                        len++;
		nl++;

		if (debug>4)
			printf("Parsed authfile line number %d: %s\n",nl,line);
		switch (line[0]){
			case 'l':
				login=xmalloc(strlen(line+1));
				strcpy(login,line+1);
			break;
			case 'p':
				password=xmalloc(strlen(line+1));
				strcpy(password,line+1);
			break;
			case 'c':
				cert=xmalloc(strlen(line+1));
				strcpy(cert,line+1);
			break;
			case 'd':
				default_mailbox=xmalloc(strlen(line+1));
                                strcpy(default_mailbox,line+1);
			break;
			default:
				die("Unknown option on line %d: %c",nl,line[0]);
		}
	}
}

int
main(int argc, char **argv)
{
	int err, opt;
	struct sockaddr_in sa;
	char* servname, * config_filename, * authdata, * home=NULL;
	struct hostent* hp;
	char* usage = "brum [-s] [-v] [-c config_file] [-a authdata_file] [-f] server port\n"
	"Note: -vv and more stand for detailed debug output displaying your passwords on the screen!";
	int option_index;
	static struct option long_options[] = {
        	{"i-am-bruce-schneier", 0, 0, 0},
		{0, 0, 0, 0}
	};

	config_filename=authdata=NULL;
	option_index = 0;
	while ((opt = getopt_long(argc, argv, "sfvc:a:",long_options, &option_index)) != -1) {
                switch (opt) {
		case 0:
			discard_security=1;
			printf("When God needs a new secure certificate, he uses Bruce Schneier as the signing authority.\n\n");	
			break;
                case 'c':
                        config_filename=xmalloc(strlen(optarg)+1);
			strcpy(config_filename,optarg);
                        break;
                case 'a':
                        authdata=xmalloc(strlen(optarg)+1);
			strcpy(authdata,optarg);
                        break;
                case 'v':
                        debug++;
                        break;
		case 'f':
			see_from++;
			break;
		case 's':
			secure++;
			break;
                default:
                        die(usage);
                }
        }

	/* read port and server (TODO: make it configurable in a file) */
	if (argc<3)
		die(usage);
	servname=argv[optind];
	port=atoi(argv[1+optind]);

		/* we need default path */
	if (!config_filename || !authdata){
		home=getenv("HOME");
		if (!home)
			die("Cannot get path to home");
	}
	if (!config_filename){
		char* default_name=".brum";
		config_filename=xmalloc(strlen(default_name)+strlen(home)+1);
		sprintf(config_filename,"%s/%s",home,default_name);	
	}
	if (!authdata){
		char* default_name=".brumauth";
		authdata=xmalloc(strlen(default_name)+strlen(home)+1);
		sprintf(authdata,"%s/%s",home,default_name);	
	}

	/* cope with configuration */
	parse_authfile(authdata);
	parse_config_file(config_filename);

	if (debug>4)
		show_recipe_list(&recipe_list);	

	/* prepare connection */
	hp=gethostbyname(servname);
	if (!hp)
		die(strerror(errno));
	sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock<0)
		die(strerror(errno));
	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	memcpy(&sa.sin_addr, hp->h_addr, 4);
	sa.sin_port = htons(port);
	if (connect(sock, (struct sockaddr *) &sa, sizeof(sa)) < 0)
		die(strerror(errno));

	if (secure){
		tls_init();
		gnutls_init(&sess, GNUTLS_CLIENT);
		gnutls_set_default_priority(sess);
		gnutls_credentials_set(sess, GNUTLS_CRD_CERTIFICATE, cert_cred);
		gnutls_transport_set_ptr(sess, (gnutls_transport_ptr_t) (intptr_t) sock);
		err = gnutls_handshake(sess); TLS_CHECK(gnutls_handshake);
	} 

	int mode = fcntl(sock, F_GETFL);
	if (mode < 0 || fcntl(sock, F_SETFL, mode | O_NONBLOCK) < 0)
		die("Cannot set non-blocking mode: %m");

	if (debug>4)
		printf("Connected\n");

	/* we can start to work with messages, init */
	init_cmds_storage();
	curtag=0x2a;	

	if (!login || !password)
		die("Login or password not supplied");
	authenticate(login,password);
	select_mailbox("INBOX");
	fetch_inbox();

	if (!default_mailbox)
		die("Do not have default mailbox");
	parse_inbox();
	apply_rules();
	move_messages();

	/* FIXME: move to separate function, see fetch_inbox */
	logout();
	xclose();

	return 0;
}
